/* 
 * mpeg2ether.c
 * 
 * Copyright (c) 2007 Koen De Vleeschauwer. 
 * 
 * Redistribution and use in source, netlist, binary and silicon forms, 
 * with or without modification, are permitted provided that the following conditions are met:
 * 
 *    1. Redistributions of source code must retain the above copyright notice, 
 *       this list of conditions and the following disclaimer.
 *    2. Redistributions in binary form must reproduce the above copyright notice, 
 *       this list of conditions and the following disclaimer in the documentation 
 *       and/or other materials provided with the distribution.
 *    3. Neither the name of the author nor the names of its contributors may be used to endorse 
 *       or promote products derived from this software without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND 
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 * ARE DISCLAIMED. IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE 
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL 
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS 
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) 
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT 
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY 
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF 
 * SUCH DAMAGE.
 */

/*
 * mpeg2ether.c
 * read/write mpeg2 decoder registers, playback mpeg2 video
 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <errno.h>
#include <getopt.h>
#include <math.h>
#include <pgm.h>
#include <fcntl.h>

#include <netdb.h>
#include <arpa/inet.h>
#include <netinet/in.h>
#include <sys/types.h>
#include <sys/socket.h>

/*
 * FIRMWARE_VERSION needs to correspond with `define VERSION in regfile.v
 * FIRMWARE_VERSION is used to check the correct firmware version is loaded.
 */
#define FIRMWARE_VERSION 12

#define UDP_PORT_MPEG2_CMD 16384

#define MPEG2_CMD_RESET 1
#define MPEG2_CMD_WRITE 2
#define MPEG2_CMD_READ  3

#define DISP_VIDEO 0
#define DISP_ALPHA 2
#define DISP_OSD   4
#define DISP_BLINK 6
#define DISP_ATTEN 0
#define DISP_BLEND 1

#define REG_RD_VERSION        0x0
#define REG_RD_STATUS         0x1
#define REG_RD_SIZE           0x2
#define REG_RD_DISP_SIZE      0x3
#define REG_RD_FRAME_RATE     0x4
#define REG_RD_TESTPOINT      0xf

#define REG_WR_STREAM         0x0
#define REG_WR_HOR            0x1
#define REG_WR_HOR_SYNC       0x2
#define REG_WR_VER            0x3
#define REG_WR_VER_SYNC       0x4
#define REG_WR_VID_MODE       0x5
#define REG_WR_CLT_YUVM       0x6
#define REG_WR_CLT_ADDR       0x7
#define REG_WR_OSD_DTA_HIGH   0x8
#define REG_WR_OSD_DTA_LOW    0x9
#define REG_WR_OSD_ADDR       0xa
#define REG_WR_TRICK          0xb
#define REG_WR_DVI            0xe
#define REG_WR_TESTPOINT      0xf

#define MAX_ROWS              1080
#define MAX_COLS              1920

#define MAX_PAYLOAD           1400

static int verbose = 0;
static int help = 0;
static float dotclock = 27.0;
static int colorbars = 0;
static int hor_resolution = 639;
static int hor_length = 799;
static int hor_sync_start = 655;
static int hor_sync_end = 751;
static int ver_resolution = 479;
static int ver_length = 524;
static int ver_sync_start = 489;
static int ver_sync_end = 491;
static int hor_halfline = 319;
static int clip_display = 0;
static int pixel_repeat = 0;
static int interlaced = 0;
static int deinterlace = 0;
static int repeat_frame = 0;
static int persistence = 0;
static int source_select = 0;
static int flush_vbuf = 0;
static int osd_enable = 0;
static int watchdog_timer = 127;	/* 0.5 second */
static int testpoint_sel = 0;
static int reg_data = 0;
static int reg_read = -1;
static int reg_write = -1;
static char *hostname = NULL;	/* fpga board hostname */
static int write_regs = 0;
static int write_stream_reg = 0;
static int write_dcm_reg = 0;
static char *osd_clt_fname = NULL;	/* osd color lookup table filename */
static char *osd_bitmap_fname = NULL;	/* osd portable graymap filename */
static int horizontal_size;
static int vertical_size;
static int osd_wr_full;

int osd_update = 0;		/* reduce verbose output during osd updates */
int osd_bitmap[MAX_COLS][MAX_ROWS];

int sock = -1;
struct sockaddr_in saddr_in;

typedef struct mpeg2_cmd {
    unsigned char cmd;
    unsigned char reg_addr;
    unsigned char reg_data[4];
} mpeg2_cmd;

typedef struct rx_pkt {
    unsigned char mpeg2_cmd;
    unsigned char mpeg2_reg;
    unsigned char mpeg2_dta[4];
    char filler[MAX_PAYLOAD];
} rx_pkt;

rx_pkt *rx_packet = NULL;

void explain()
{
    printf("Options:\n");
    printf(" --verbose\n");
    printf(" --brief\n");
    printf(" --help\n");
    printf(" --hostname <ip_address>\n");

    printf(" --reset\n");
    printf(" --ping\n");

    printf(" --dot_clock <float>\n");
    printf(" --color_bars <number>\n");
    printf("   force color bars test pattern, e.g. --dot_clock 27 --color_bars 1\n");
    printf(" --hor_resolution <number>\n");
    printf(" --hor_length <number>\n");
    printf(" --hor_sync_start <number>\n");
    printf(" --hor_sync_end <number>\n");
    printf(" --ver_resolution <number>\n");
    printf(" --ver_length <number>\n");
    printf(" --ver_sync_start <number>\n");
    printf(" --ver_sync_end <number>\n");
    printf(" --hor_halfline <number>\n");
    printf(" --clip_display <number>\n");
    printf(" --pixel_repeat <number>\n");
    printf(" --interlaced <number>\n");
    printf(" --deinterlace <number>\n");
    printf(" --repeat_frame <number>\n");
    printf("   image repeats <number> + 1 times; 31 = forever\n");
    printf(" --persistence <number>\n");
    printf("   last image persists. e.g. --persistence 1\n");
    printf(" --source_select <number>\n");
    printf(" --flush_vbuf <number>\n");
    printf(" --watchdog_timer <number>\n");

    printf(" --testpoint_sel <number>\n");

    printf(" --enable-osd\n");
    printf(" --disable-osd\n");
    printf(" --osd_clt <filename>\n");
    printf("   load on-screen dislay color lookup table\n");
    printf(" --osd_bitmap <filename>\n");
    printf("   load on-screen dislay bitmap (pgm format)\n");

    printf(" --read <register>\n");
    printf(" --data <number> --write <register>\n");
}

int err_exit(char *msg)
{
    if (errno == 0)
	printf("%s\n", msg);
    else
	perror(msg);
    exit(-1);
}


unsigned int strtou(const char *str, char **endptr, int base)
{
    long l;
    unsigned int u;
    l = strtoul(str, endptr, base);
    u = l;
    return u;
}

unsigned int getbits(int i, int pos, int len)
{
    return (i >> (pos - len + 1)) & ~(~0 << len);
}

unsigned int setbits(int i, int pos, int len)
{
    return (i & ~(~0 << len)) << (pos - len + 1);
}

/*
 * sockets
 */

void get_socket()
{
    struct hostent *hep;

    if (sock < 0) {
	if ((sock = socket(PF_INET, SOCK_DGRAM, IPPROTO_UDP)) < 0)
	    err_exit("could not open udp socket");

	if (hostname == NULL)
	    err_exit("--hostname required");

	hep = gethostbyname(hostname);

	if ((hep == NULL) || (*hep->h_addr_list == NULL))
	    err_exit("fpga ip address not found");
	if (hep->h_addrtype != AF_INET)
	    err_exit("not an ipv4 address");

	memset(&saddr_in, 0, sizeof(saddr_in));
	memcpy(&saddr_in.sin_addr.s_addr, *hep->h_addr_list, hep->h_length);

	saddr_in.sin_family = AF_INET;
	saddr_in.sin_port = htons(UDP_PORT_MPEG2_CMD);
	printf("Using address %s\n", inet_ntoa(saddr_in.sin_addr));
    }

    if (rx_packet == NULL)
	rx_packet = (rx_pkt *) malloc(MAX_PAYLOAD);
    if (rx_packet == NULL)
	err_exit("Could not allocate receive packet memory");
}

void send_packet(char *payload, int payload_len)
{
    get_socket();
    if (sendto(sock, payload, payload_len, 0, (struct sockaddr *) &saddr_in, (socklen_t) sizeof(saddr_in)) < 0)
	err_exit("sendto error");
}

int receive_packet(char *payload, int payload_len)
{
    return recvfrom(sock, payload, payload_len, 0, NULL, NULL);	// XXX need timeout in case packet is dropped
}

/*
 * decode received packet 
 */

void receive_ack()
{
    int recv_len;
    unsigned int dta;
    int version;

    /* receive response */
    recv_len = receive_packet((char *) rx_packet, MAX_PAYLOAD);
    //usleep (200000);
    if (recv_len > 4) {
	memcpy(&dta, rx_packet->mpeg2_dta, 4);
	if (rx_packet->mpeg2_cmd == 128) {
	    /* ack of a mpeg2 ts or es  packet */
	} else {
	    if (!osd_update)
		printf(" Received packet, cmd: 0x%02x", rx_packet->mpeg2_cmd);
	    if (rx_packet->mpeg2_cmd == (MPEG2_CMD_RESET + 128)) {
		if (!osd_update)
		    printf(" (reset ack)\n");
	    } else if (rx_packet->mpeg2_cmd == (MPEG2_CMD_WRITE + 128)) {
		if (!osd_update)
		    printf(" (write ack)\n");
	    } else if (rx_packet->mpeg2_cmd == (MPEG2_CMD_READ + 128)) {
		if (!osd_update)
		    printf(" (read ack), reg: %02x dta: %08x\n", rx_packet->mpeg2_reg, dta);	// better use hton() and ntoh()
		switch (rx_packet->mpeg2_reg) {
		case 0:
		    if (!osd_update) {
			version = getbits(dta, 15, 16);
			printf(" version: %0i\n", version);
			if (version != FIRMWARE_VERSION)
			    err_exit("software/firmware version mismatch");
		    }
		    break;
		case 1:
		    if (!osd_update) {
			printf(" matrix_coefficients: %u\n", getbits(dta, 15, 8));
			printf(" watchdog_status: %u\n", getbits(dta, 7, 1));
			printf(" osd_wr_en: %u\n", getbits(dta, 6, 1));
			printf(" osd_wr_ack: %u\n", getbits(dta, 5, 1));
			printf(" osd_wr_full: %u\n", getbits(dta, 4, 1));
			printf(" picture_hdr: %u\n", getbits(dta, 3, 1));
			printf(" frame_end: %u\n", getbits(dta, 2, 1));
			printf(" video_ch: %u\n", getbits(dta, 1, 1));
			printf(" error: %u\n", getbits(dta, 0, 1));
		    }
		    osd_wr_full = getbits(dta, 4, 1);
		    break;
		case 2:
		    if (!osd_update) {
			printf(" horizontal_size: %u\n", getbits(dta, 29, 14));
			printf(" vertical_size: %u\n", getbits(dta, 13, 14));
		    }
		    horizontal_size = getbits(dta, 29, 14);
		    vertical_size = getbits(dta, 13, 14);
		    break;
		case 3:
		    if (!osd_update) {
			printf(" display_horizontal_size: %u\n", getbits(dta, 29, 14));
			printf(" display_vertical_size: %u\n", getbits(dta, 13, 14));
		    }
		    break;
		case 4:
		    if (!osd_update) {
			printf(" aspect_ratio_information: %u\n", getbits(dta, 15, 4));
			printf(" progressive_sequence: %u\n", getbits(dta, 11, 1));
			printf(" frame_rate_extension_d: %u\n", getbits(dta, 10, 5));
			printf(" frame_rate_extension_n: %u\n", getbits(dta, 5, 2));
			printf(" frame_rate_code: %u\n", getbits(dta, 3, 4));
		    }
		    break;
		}
	    } else if (!osd_update)
		printf(" unknown packet type %i\n", rx_packet->mpeg2_cmd);
	}
    }
}

/*
 * Register access
 */

void write_reset()
{
    mpeg2_cmd reset_cmd;

    reset_cmd.cmd = MPEG2_CMD_RESET;
    reset_cmd.reg_addr = 0;
    memset(&reset_cmd.reg_data, 0, sizeof(reset_cmd.reg_data));

    if (verbose)
	printf("reset\n");
    send_packet((char *) &reset_cmd, sizeof(reset_cmd));
    receive_ack();
}

void write_register(int regno, int regdta)
{
    int regdta_n;
    mpeg2_cmd write_cmd;

    regdta_n = htonl(regdta);
    write_cmd.cmd = MPEG2_CMD_WRITE;
    write_cmd.reg_addr = regno;
    memcpy(write_cmd.reg_data, &regdta_n, sizeof(regdta_n));

    if (verbose && !osd_update)
	printf("write_register: reg 0x%02x <= 0x%08x\n", regno, regdta);

    send_packet((char *) &write_cmd, sizeof(write_cmd));
    receive_ack();
}

void read_register(int regno)
{
    mpeg2_cmd read_cmd;

    read_cmd.cmd = MPEG2_CMD_READ;
    read_cmd.reg_addr = regno;
    memset(&read_cmd.reg_data, 0, sizeof(read_cmd.reg_data));

    if (verbose && !osd_update)
	printf("read_register:  reg 0x%02x\n", regno);

    send_packet((char *) &read_cmd, sizeof(read_cmd));
    receive_ack();
}


/*
 * DVI dot clock:
 * dcm frequency synthesizer registers - see comments in dotclock.v
 */

void write_dcm()
{
    int multiplier, divider;
    int high_freq_mode;
    float fin, fclkfx, fvco, fout;	/* frequencies in MHz */

    float fdesired, fobtained, ferror;
    int m, d, hfm;
    int lowfreq;		// for chrontel ch7301

    if ((dotclock < 25.0) || (dotclock > 90.0)) {
	printf("Supported dot clock range: 25..90 MHz\n");
	printf("For frequencies below 25 MHz, double the frequency and use --pixel_repeat 1\n");
	exit(0);
    }

    /*
     * Loop over all possible multiplier/divider pairs, 
     * select the multiplier/divider pair which comes closest to the desired frequency
     */

    fdesired = dotclock;
    ferror = 700.0;
    m = 0;
    d = 0;
    hfm = 0;

    fin = 100.0;
    for (multiplier = 2; multiplier <= 32; multiplier++)
	for (divider = 2; divider <= 32; divider++) {
	    fclkfx = fin * (float) multiplier / (float) divider;
	    fvco = fclkfx * 4.0;
	    fout = fvco / 16.0;
	    high_freq_mode = (fclkfx >= 160.0);
	    if ((32.0 <= fclkfx) && (fclkfx <= 400) && (400.0 <= fvco) && (fvco <= 1440.0) && (3.125 <= fout <= 700.0)) {
		// printf (" %7.2f MHz M: %2i D: %2i high: %i\n", fout, multiplier, divider, high_freq_mode);
		if ((fabs(fout - fdesired) < ferror) || ((fabs(fout - fdesired) == ferror) && ((multiplier + divider) < (m + d)))) {
		    fobtained = fout;
		    ferror = fabs(fout - fdesired);
		    m = multiplier;
		    d = divider;
		    hfm = high_freq_mode;
		    // printf (" better\n");
		}
	    }
	}

    lowfreq = (fobtained <= 65.0);

    if (verbose)
	printf("dotclock ftarget = %6.2f fout = %6.2f MHz multiplier: %2i divider: %2i high frequency mode: %1i ch7301 lowfreq: %1i ch7301 colorbars: %1i\n", fdesired, fobtained, m, d, hfm, lowfreq, colorbars);
    write_register(REG_WR_DVI, setbits(colorbars, 19, 1) | setbits(lowfreq, 18, 1) | setbits(hfm, 17, 1) | setbits(hfm, 16, 1) | setbits(m - 1, 15, 8) | setbits(d - 1, 7, 8));
}

void write_registers()
{
    if (write_dcm_reg)
	write_dcm();
    if (write_stream_reg)
	write_register(REG_WR_STREAM, setbits(watchdog_timer, 15, 8) | setbits(osd_enable, 3, 1));
    if (!write_regs)
	return;
    if (verbose)
	printf("write_registers\n");
    write_register(REG_WR_HOR, setbits(hor_resolution, 27, 12) | setbits(hor_length, 11, 12));
    write_register(REG_WR_HOR_SYNC, setbits(hor_sync_start, 27, 12) | setbits(hor_sync_end, 11, 12));
    write_register(REG_WR_VER, setbits(ver_resolution, 27, 12) | setbits(ver_length, 11, 12));
    write_register(REG_WR_VER_SYNC, setbits(ver_sync_start, 27, 12) | setbits(ver_sync_end, 11, 12));
    write_register(REG_WR_VID_MODE, setbits(hor_halfline, 27, 12) | setbits(clip_display, 2, 1) | setbits(pixel_repeat, 1, 1) | setbits(interlaced, 0, 1));
    write_register(REG_WR_TRICK, setbits(deinterlace, 10, 1) | setbits(repeat_frame, 9, 5) | setbits(persistence, 4, 1) | setbits(source_select, 3, 3) | setbits(flush_vbuf, 0, 1));
    write_register(REG_WR_TESTPOINT, setbits(testpoint_sel, 3, 4));
    write_regs = 0;
}

void write_osd_clt_entry(int addr, int y, int u, int v, int mode, int alpha)
{
    int osd_clt_yuvm;
    int osd_clt_addr;
    osd_clt_yuvm = setbits(y, 31, 8) | setbits(u, 23, 8) | setbits(v, 15, 8) | setbits(mode, 7, 3) | setbits(alpha, 4, 5);
    osd_clt_addr = setbits(addr, 7, 8);
    write_register(REG_WR_CLT_YUVM, osd_clt_yuvm);
    write_register(REG_WR_CLT_ADDR, osd_clt_addr);
}

void write_osd_clt()
{
    int i;
    FILE *fd;
    int addr, y, u, v, mode, alpha;
    char line[256];
    char mode_str[16];

    if (osd_clt_fname == NULL)
	return;
    /* clear osd clt */
    osd_update = 1;
    for (i = 0; i < 256; i++)
	write_osd_clt_entry(i, 0, 0, 0, DISP_VIDEO, 16);
    /* read osd clt file */
    if ((fd = fopen(osd_clt_fname, "r")) == NULL)
	err_exit("Could not open clt file");
    printf("Loading osd clt\n");
    while (fgets(line, sizeof(line), fd) != NULL) {
	if (sscanf(line, "%d %d %d %d %s %d", &addr, &y, &u, &v, mode_str, &alpha) == 6) {
	    if (strcmp(mode_str, "video") == 0)
		mode = DISP_VIDEO;
	    else if (strcmp(mode_str, "alpha_atten") == 0)
		mode = DISP_ALPHA | DISP_ATTEN;
	    else if (strcmp(mode_str, "alpha_blend") == 0)
		mode = DISP_ALPHA | DISP_BLEND;
	    else if (strcmp(mode_str, "osd") == 0)
		mode = DISP_OSD;
	    else if (strcmp(mode_str, "blink_atten") == 0)
		mode = DISP_BLINK | DISP_ATTEN;
	    else if (strcmp(mode_str, "blink_blend") == 0)
		mode = DISP_BLINK | DISP_BLEND;
	    else
		mode = 0;
	    if (verbose)
		printf("osd clt addr: %d y: %d u: %d v: %d mode: %d alpha: %d\n", addr, y, u, v, mode, alpha);
	    write_osd_clt_entry(addr, y, u, v, mode, alpha);
	}
    }
    fclose(fd);
    osd_update = 0;
}

write_osd_bitmap()
{
    FILE *fd;
    int x, y;
    gray **pgm_bitmap;
    gray max;
    int cols, rows;
    int osd_dta_high;
    int osd_dta_low;
    int osd_frame;
    int osd_comp;
    int osd_addr_x;
    int osd_addr_y;
    int osd_addr;

    if (osd_bitmap_fname == NULL)
	return;
/* 
 * Before writing to the osd, horizontal_size and vertical_size need to be defined.
 * Write a sequence header to the mpeg2 decoder to initalize horizontal_size and vertical_size.
 */
    read_register(REG_RD_SIZE);
    if ((horizontal_size == 0) || (vertical_size == 0))
	err_exit("video size not set");
    /* initialize bitmap */
    for (y = 0; y < MAX_ROWS; y++)
	for (x = 0; x < MAX_COLS; x++)
	    osd_bitmap[x][y] = 0;
    /* read pgm bitmap using netpbm */
    if ((fd = fopen(osd_bitmap_fname, "r")) == NULL)
	err_exit("Could not open pgm bitmap");
    pgm_bitmap = pgm_readpgm(fd, &cols, &rows, &max);
    /* copy pgm bitmap */
    for (y = 0; (y < rows) && (y < MAX_ROWS); y++)
	for (x = 0; (x < cols) && (x < MAX_COLS); x++)
	    osd_bitmap[x][y] = pgm_bitmap[y][x];
    pgm_freearray(pgm_bitmap, rows);
    fclose(fd);
    /* write to fpga */
    printf("Loading osd bitmap\n");
    osd_update = 1;
    for (y = 0; y < vertical_size; y++)
	for (x = 0; x < horizontal_size; x += 8) {
	    osd_dta_high = setbits(osd_bitmap[x][y], 31, 8) | setbits(osd_bitmap[x + 1][y], 23, 8) | setbits(osd_bitmap[x + 2][y], 15, 8) | setbits(osd_bitmap[x + 3][y], 7, 8);
	    osd_dta_low = setbits(osd_bitmap[x + 4][y], 31, 8) | setbits(osd_bitmap[x + 5][y], 23, 8) | setbits(osd_bitmap[x + 6][y], 15, 8) | setbits(osd_bitmap[x + 7][y], 7, 8);
	    osd_frame = 4;
	    osd_comp = 0;
	    osd_addr_x = x;
	    osd_addr_y = y;
	    osd_addr = setbits(osd_frame, 31, 3) | setbits(osd_comp, 2, 28) | setbits(osd_addr_x, 26, 11) | setbits(osd_addr_y, 10, 11);
	    write_register(REG_WR_OSD_DTA_HIGH, osd_dta_high);
	    write_register(REG_WR_OSD_DTA_LOW, osd_dta_low);
	    read_register(REG_RD_STATUS);
	    while (osd_wr_full) {
		sleep(1);	// XXX Else polling might flood the lan
		read_register(REG_RD_STATUS);
	    }
	    write_register(REG_WR_OSD_ADDR, osd_addr);
	}
    /* repeat last write 32 times to flush osd write fifo */
    for (y = 0; y < 31; y++) {
	while (osd_wr_full) {
	    sleep(1);		// XXX Else polling might flood the lan
	    read_register(REG_RD_STATUS);
	}
	write_register(REG_WR_OSD_ADDR, osd_addr);
    }

    if (!verbose)
	printf("\n");
    osd_update = 0;
}

/*
 * Main
 */

int main(argc, argv)
int argc;
char **argv;
{
    int c;

    pgm_init(&argc, argv);

    while (1) {
	static struct option long_options[] = {
	    {"verbose", no_argument, &verbose, 1},
	    {"brief", no_argument, &verbose, 0},
	    {"help", no_argument, 0, 'h'},
	    {"hostname", required_argument, 0, 'B'},
	    {"dot_clock", required_argument, 0, 'C'},
	    {"color_bars", required_argument, 0, 'D'},
	    {"hor_resolution", required_argument, 0, 'b'},
	    {"hor_length", required_argument, 0, 'c'},
	    {"hor_sync_start", required_argument, 0, 'e'},
	    {"hor_sync_end", required_argument, 0, 'g'},
	    {"ver_resolution", required_argument, 0, 'y'},
	    {"ver_length", required_argument, 0, 'i'},
	    {"ver_sync_start", required_argument, 0, 'j'},
	    {"ver_sync_end", required_argument, 0, 'k'},
	    {"hor_halfline", required_argument, 0, 'l'},
	    {"clip_display", required_argument, 0, 'E'},
	    {"pixel_repeat", required_argument, 0, 'm'},
	    {"interlaced", required_argument, 0, 'n'},
	    {"deinterlace", required_argument, 0, 'p'},
	    {"repeat_frame", required_argument, 0, 'o'},
	    {"persistence", required_argument, 0, 'x'},
	    {"source_select", required_argument, 0, 'q'},
	    {"flush_vbuf", required_argument, 0, 's'},
	    {"watchdog_timer", required_argument, 0, 'z'},
	    {"testpoint_sel", required_argument, 0, 'v'},
	    {"osd_enable", no_argument, 0, 'F'},
	    {"osd_disable", no_argument, 0, 'G'},
	    {"osd_clt", required_argument, 0, 'H'},
	    {"osd_bitmap", required_argument, 0, 'I'},
	    {"data", required_argument, 0, 'd'},
	    {"read", required_argument, 0, 'r'},
	    {"write", required_argument, 0, 'w'},
	    {"reset", no_argument, 0, 'Z'},
	    {0, 0, 0, 0}
	};
	/* |getopt_long| stores the option index here. */
	int option_index = 0;

	c = getopt_long(argc, argv, "b:c:d:e:g:hi:j:k:l:m:n:o:p:q:r:s:v:w:x:y:z:B:C:D:EFGH:I:Z", long_options, &option_index);

	/* Detect the end of the options. */
	if (c == -1)
	    break;

	switch (c) {
	case 0:
	    /* If this option set a flag, do nothing else now. */
	    if (long_options[option_index].flag != 0)
		break;
	    printf("option %s", long_options[option_index].name);
	    if (optarg)
		printf(" with arg %s", optarg);
	    printf("\n");
	    break;

	case 'h':
	    explain();
	    break;

	case 'b':
	    hor_resolution = strtou(optarg, NULL, 0);
	    write_regs = 1;
	    break;

	case 'c':
	    hor_length = strtou(optarg, NULL, 0);
	    write_regs = 1;
	    break;

	case 'e':
	    hor_sync_start = strtou(optarg, NULL, 0);
	    write_regs = 1;
	    break;

	case 'g':
	    hor_sync_end = strtou(optarg, NULL, 0);
	    write_regs = 1;
	    break;

	case 'y':
	    ver_resolution = strtou(optarg, NULL, 0);
	    write_regs = 1;
	    break;

	case 'i':
	    ver_length = strtou(optarg, NULL, 0);
	    write_regs = 1;
	    break;

	case 'j':
	    ver_sync_start = strtou(optarg, NULL, 0);
	    write_regs = 1;
	    break;

	case 'k':
	    ver_sync_end = strtou(optarg, NULL, 0);
	    write_regs = 1;
	    break;

	case 'l':
	    hor_halfline = strtou(optarg, NULL, 0);
	    write_regs = 1;
	    break;

	case 'E':
	    clip_display = strtou(optarg, NULL, 0);
	    write_regs = 1;
	    break;

	case 'm':
	    pixel_repeat = strtou(optarg, NULL, 0);
	    write_regs = 1;
	    break;

	case 'n':
	    interlaced = strtou(optarg, NULL, 0);
	    write_regs = 1;
	    break;

	case 'p':
	    deinterlace = strtou(optarg, NULL, 0);
	    write_regs = 1;
	    break;

	case 'o':
	    repeat_frame = strtou(optarg, NULL, 0);
	    write_regs = 1;
	    break;

	case 'x':
	    persistence = strtou(optarg, NULL, 0);
	    write_regs = 1;
	    break;

	case 'q':
	    source_select = strtou(optarg, NULL, 0);
	    write_regs = 1;
	    break;

	case 's':
	    flush_vbuf = strtou(optarg, NULL, 0);
	    write_regs = 1;
	    break;

	case 'z':
	    watchdog_timer = strtou(optarg, NULL, 0);
	    write_stream_reg = 1;
	    break;

	case 'v':
	    testpoint_sel = strtou(optarg, NULL, 0);
	    write_regs = 1;
	    break;

	case 'd':
	    write_registers();
	    reg_data = strtou(optarg, NULL, 0);
	    break;

	case 'r':
	    write_registers();
	    reg_read = strtou(optarg, NULL, 0);
	    read_register(reg_read);
	    break;

	case 'w':
	    write_registers();
	    reg_write = strtou(optarg, NULL, 0);
	    write_register(reg_write, reg_data);
	    break;

	case 'B':
	    hostname = optarg;
	    break;

	case 'C':
	    sscanf(optarg, "%f", &dotclock);
	    write_dcm_reg = 1;
	    break;

	case 'D':
	    colorbars = (strtou(optarg, NULL, 0) != 0);
	    write_dcm_reg = 1;
	    break;

	case 'Z':
	    write_registers();
	    write_reset();
	    break;

	case 'F':
	    osd_enable = 1;
	    write_stream_reg = 1;
	    break;

	case 'G':
	    osd_enable = 0;
	    write_stream_reg = 1;
	    break;

	case 'H':
	    write_registers();
	    osd_clt_fname = optarg;
	    write_osd_clt();
	    break;

	case 'I':
	    write_registers();
	    osd_bitmap_fname = optarg;
	    write_osd_bitmap();
	    break;

	case '?':
	    /* |getopt_long| already printed an error message. */
	    explain();
	    break;

	default:
	    abort();
	}
    }

    write_registers();

    if (help)
	explain();
    /* Print any remaining command line arguments (not options). */
    if (optind < argc) {
	printf("Unknown options: ");
	while (optind < argc)
	    printf("%s ", argv[optind++]);
	putchar('\n');
    }

    exit(0);
}

/* not truncated */
